import tensorflow as tf
from KENN2.layers.residual.KnowledgeEnhancer import KnowledgeEnhancer


class Kenn(tf.keras.layers.Layer):

    def __init__(self, predicates, clauses, activation=lambda x: x, initial_clause_weight=0.5, **kwargs):
        """Initialize the knowledge base.

        :param predicates: a list of predicates names
        :param clauses: a list of constraints. Each constraint is a string on the form:
        clause_weight:clause

        The clause_weight should be either a real number (in such a case this value is fixed) or an underscore
        (in this case the weight will be a tensorflow variable and learned during training).

        The clause must be represented as a list of literals separated by commas (that represent disjunctions).
        Negation must specified by adding the letter 'n' before the predicate name.

        An example:
           _:nDog,Animal

        """

        super(Kenn, self).__init__(**kwargs)

        self.predicates = predicates
        self.clauses = clauses
        self.activation = activation
        self.initial_clause_weight = initial_clause_weight
        self.knowledge_enhancer = None

    def build(self, input_shape):
        """Build the layer

        :param input_shape: the input shape
        """

        self.knowledge_enhancer = KnowledgeEnhancer(self.predicates, self.clauses, self.initial_clause_weight)

        super(Kenn, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Improve the satisfaction level of a set of clauses.

        :param inputs: the tensor containing predicates' pre-activation values for many entities
        :return: final preactivations"""

        deltas = self.knowledge_enhancer(inputs)

        return self.activation(inputs + deltas)

    def get_config(self):
        config = super(Kenn, self).get_config()
        config.update({'predicates':self.predicates})
        config.update({'clauses':self.clauses})
        config.update({'activation':self.activation})
        config.update({'initial_clause_weight':self.initial_clause_weight})
        # config['output_size'] =  # say self. _output_size  if you store the argument in __init__
        return config


import tensorflow as tf
import numpy as np

from KENN2.layers.RangeConstraint import RangeConstraint


class ClauseEnhancer(tf.keras.layers.Layer):
    """Clause Enhancer layer
    """

    def __init__(self, available_predicates, clause_string, initial_clause_weight):
        """Initialize the clause.
        :param available_predicates: the list of all possible literals in a clause
        :param clause_string: a string representing a conjunction of literals. The format should be:
        clause_weight:clause
        The clause_weight should be either a real number (in such a case this sign is fixed) or an underscore
        (in this case the weight will be a tensorflow variable and learned during training).
        The clause must be represented as a list of literals separated by commas (that represent disjunctions).
        Negation must specified by adding the letter 'n' before the predicate name.
        An example:
           _:nDog,Animal
        :param initial_clause_weight: the initial sign to the clause weight. Used if the clause weight is learned.
        """

        super(ClauseEnhancer, self).__init__()
        string = clause_string.split(':')

        self.original_string = string[1]
        self.string = string[1].replace(
            ',', 'v').replace('(', '').replace(')', '')

        if string[0] == '_':
            self.initial_weight = initial_clause_weight
            self.hard_clause = False
        else:
            self.initial_weight = int(string[0])
            self.hard_clause = True

        self.hard_clause = string[0] != '_'

        literals_list = string[1].split(',')
        self.number_of_literals = len(literals_list)

        self.gather_literal_indices = []
        self.scatter_literal_indices = []
        signs = []

        for literal in literals_list:
            sign = 1
            if literal[0] == 'n':
                sign = -1
                literal = literal[1:]

            literal_index = available_predicates.index(literal)
            self.gather_literal_indices.append(literal_index)
            self.scatter_literal_indices.append([literal_index])
            signs.append(sign)

        self.signs = np.array(signs, dtype=np.float32)
        self.clause_weight = None

    def build(self, input_shape):
        """Build the layer
        :param input_shape: the input shape
        """

        self.clause_weight = self.add_weight(
            name='kernel',
            shape=(1, 1),
            initializer=tf.keras.initializers.Constant(
                value=self.initial_weight),
            constraint=RangeConstraint(),
            trainable=not self.hard_clause)

        super(ClauseEnhancer, self).build(input_shape)

    def grounded_clause(self, inputs):
        """Find the grounding of the clause
        :param inputs: the tensor containing predicates' pre activations for many objects
        :return: the grounded clause (a tensor with literals truth values)
        """

        selected_predicates = tf.gather(
            inputs, self.gather_literal_indices, axis=1)
        clause_matrix = selected_predicates * self.signs

        return clause_matrix

    def call(self, inputs, **kwargs):
        """Improve the satisfaction level of the clause.
        :param inputs: the tensor containing predicates' pre-activation values for many entities
        :return: delta vector to be summed to the original pre-activation tensor to obtain an higher satisfaction of \
        the clause"""

        clause_matrix = self.grounded_clause(inputs)

        delta = self.signs * tf.nn.softmax(clause_matrix) * self.clause_weight

        scattered_delta = tf.scatter_nd(self.scatter_literal_indices, tf.transpose(
            delta), tf.reverse(tf.shape(inputs), [0]))

        return tf.transpose(scattered_delta)



from KENN2.layers.residual.KnowledgeEnhancer import KnowledgeEnhancer
from KENN2.layers.Kenn import Kenn
from KENN2.layers.RelationalKENN import RelationalKENN


def unary_parser(knowledge_file, activation=lambda x: x, initial_clause_weight=0.5,**kwargs):
    """
    Takes in input the knowledge file containing only unary clauses and returns a Kenn Layer, 
    with input the predicates and clauses found in the knowledge file.

    :param knowledge_file: path of the prior knowledge file
    """
    with open(knowledge_file, 'r') as kb_file:
        predicates_string = kb_file.readline()
        kb_file.readline()
        clauses = kb_file.readlines()

    predicates = predicates_string[:-1].split(',')

    return Kenn(predicates, clauses, activation, initial_clause_weight, **kwargs)


def unary_parser_ke(knowledge_file, initial_clause_weight=0.5, **kwargs):
    """
    Takes in input the knowledge file containing only unary clauses and returns a Knowledge Enhancer layer, 
    with input the predicates and clauses found in the knowledge file.

    :param knowledge_file: path of the prior knowledge file;
    """
    with open(knowledge_file, 'r') as kb_file:
        predicates_string = kb_file.readline()
        kb_file.readline()
        clauses = kb_file.readlines()

    predicates = predicates_string[:-1].split(',')

    return KnowledgeEnhancer(predicates, clauses, initial_clause_weight, **kwargs)


def relational_parser(knowledge_file, activation=lambda x: x, initial_clause_weight=0.5, **kwargs):
    """
    Takes in input the knowledge file containing both unary and binary clauses and returns a RelationalKENN 
    Layer, with input the predicates and clauses found in the knowledge file.

    :param knowledge_file: path of the prior knowledge file;
    """
    with open(knowledge_file, 'r') as kb_file:
        unary_literals_string = kb_file.readline()
        binary_literals_string = kb_file.readline()

        kb_file.readline()
        clauses = kb_file.readlines()

    u_groundings = [u + '(x)' for u in unary_literals_string[:-1].split(',')]
    b_groundings = [u + '(x)' for u in unary_literals_string[:-1].split(',')] + \
                   [u + '(y)' for u in unary_literals_string[:-1].split(',')] + \
                   [b + '(x.y)' for b in binary_literals_string[:-1].split(',')] + \
                   [b + '(y.x)' for b in binary_literals_string[:-1].split(',')]

    unary_clauses = []
    binary_clauses = []

    reading_unary = True
    for clause in clauses:
        if clause[0] == '>':
            reading_unary = False
            continue

        if reading_unary:
            unary_clauses.append(clause)
        else:
            binary_clauses.append(clause)

    return RelationalKENN(
        u_groundings, 
        b_groundings, 
        unary_clauses, 
        binary_clauses, 
        activation, 
        initial_clause_weight)



import tensorflow as tf


class RangeConstraint(tf.keras.constraints.Constraint):
    """Represent a range constraint for the weight of ClauseEnhancer.

    """

    def __init__(self, min=0, max=500):
        super(RangeConstraint, self).__init__()
        self.min = min
        self.max = max

    def __call__(self, weigth):
        """Clip the weigth value inside the range [min, max]

        :return: the clipped weight
        """
        return tf.clip_by_value(weigth, self.min, self.max)



import tensorflow as tf
import numpy as np
from KENN2.layers.residual.KnowledgeEnhancer import KnowledgeEnhancer
from KENN2.layers.relational.Join import Join
from KENN2.layers.relational.GroupBy import GroupBy


class RelationalKENN(tf.keras.layers.Layer):

    def __init__(self, unary_predicates, binary_predicates, unary_clauses, binary_clauses, activation=lambda x: x, initial_clause_weight=0.5, **kwargs):
        """Initialize the knowledge base.

        :param unary_predicates: the list of unary predicates names
        :param binary_predicates: the list of binary predicates names
        :param unary_clauses: a list of unary clauses. Each clause is a string on the form:
        clause_weight:clause

        The clause_weight should be either a real number (in such a case this value is fixed) or an underscore
        (in this case the weight will be a tensorflow variable and learned during training).

        The clause must be represented as a list of literals separated by commas (that represent disjunctions).
        Negation must specified by adding the letter 'n' before the predicate name.

        An example:
           _:nDog,Animal

        :param binary_clauses: a list of binary clauses
        :param activation: activation function
        :param initial_clause_weight: initial value for the cluase weight (if clause is not hard)
        """

        super(RelationalKENN, self).__init__(**kwargs)

        self.unary_predicates = unary_predicates
        self.n_unary = len(unary_predicates)
        self.unary_clauses = unary_clauses
        self.binary_predicates = binary_predicates
        self.binary_clauses = binary_clauses
        self.activation = activation
        self.initial_clause_weight = initial_clause_weight

        self.unary_ke = None
        self.binary_ke = None
        self.join = None
        self.group_by = None

    def build(self, input_shape):
        if len(self.unary_clauses) != 0:
            self.unary_ke = KnowledgeEnhancer(
                self.unary_predicates, self.unary_clauses, initial_clause_weight=self.initial_clause_weight)
        if len(self.binary_clauses) != 0:
            self.binary_ke = KnowledgeEnhancer(
                self.binary_predicates, self.binary_clauses, initial_clause_weight=self.initial_clause_weight)

            self.join = Join()
            self.group_by = GroupBy(self.n_unary)

        super(RelationalKENN, self).build(input_shape)

    def call(self, unary, binary, index1, index2, **kwargs):
        """Forward step of Kenn model for relational data.

        :param unary: the tensor with unary predicates pre-activations
        :param binary: the tensor with binary predicates pre-activations
        :param index1: a vector containing the indices of the first object
        of the pair referred by binary tensor
        :param index2: a vector containing the indices of the second object
        of the pair referred by binary tensor
        """

        if len(self.unary_clauses) != 0:
            deltas_sum, deltas_u_list = self.unary_ke(unary)
            u = unary + deltas_sum
        else:
            u = unary
            deltas_u_list = tf.expand_dims(tf.zeros(unary.shape), axis=0)

        if len(self.binary_clauses) != 0 and len(binary) != 0:
            joined_matrix = self.join(u, binary, index1, index2)
            deltas_sum, deltas_b_list = self.binary_ke(joined_matrix)

            delta_up, delta_bp = self.group_by(
                u, binary, deltas_sum, index1, index2)
        else:
            delta_up = tf.zeros(u.shape)
            delta_bp = tf.zeros(binary.shape)

        return self.activation(u + delta_up), self.activation(binary + delta_bp)

    def get_config(self):
        config = super(RelationalKENN, self).get_config()
        config.update({'unary_predicates': self.unary_predicates})
        config.update({'unary_clauses': self.unary_clauses})
        config.update({'binary_predicates': self.binary_predicates})
        config.update({'binary_clauses': self.binary_clauses})
        config.update({'activation': self.activation})
        config.update({'initial_clause_weight': self.initial_clause_weight})

        return config



import tensorflow as tf
from KENN2.layers.residual.ClauseEnhancer import ClauseEnhancer


class KnowledgeEnhancer(tf.keras.layers.Layer):

    def __init__(self, predicates, clauses, initial_clause_weight=0.5, **kwargs):
        """Initialize the knowledge base.

        :param predicates: a list of predicates names
        :param clauses: a list of constraints. Each constraint is a string on the form:
        clause_weight:clause

        The clause_weight should be either a real number (in such a case this value is fixed) or an underscore
        (in this case the weight will be a tensorflow variable and learned during training).

        The clause must be represented as a list of literals separated by commas (that represent disjunctions).
        Negation must specified by adding the letter 'n' before the predicate name.

        An example:
           _:nDog,Animal

        :param initial_clause_weight: the initial sign to the clause weight. Used if the clause weight is learned.

        """

        super(KnowledgeEnhancer, self).__init__(**kwargs)
        self.predicates = predicates
        self.clauses = clauses
        self.initial_clause_weight = initial_clause_weight
        self.clause_enhancers = []

    def build(self, input_shape):
        """Build the layer

        :param input_shape: the input shape
        """

        for clause in self.clauses:
            self.clause_enhancers.append(ClauseEnhancer(self.predicates, clause[:-1], self.initial_clause_weight))

        super(KnowledgeEnhancer, self).build(input_shape)

    def call(self, inputs, **kwargs):
        """Improve the satisfaction level of a set of clauses.

        :param inputs: the tensor containing predicates' pre-activation values for many entities
        :return: final delta values"""

        #deltas_list will be the list of deltas for each clause
        # e.g. deltas_list[0] are the deltas relative to the first clause.
        deltas_list = []
        for clause in self.clause_enhancers:
            deltas_list.append(clause(inputs))

        return tf.add_n(deltas_list), deltas_list
 



import tensorflow as tf


class Join(tf.keras.layers.Layer):
    """Join layer

    """

    def __init__(self):
        """Initialize the Join layer.

        """
        super(Join, self).__init__()

    def build(self, input_shape):
        """Build the layer

        :param input_shape: the input shape
        """
        super(Join, self).build(input_shape)

    def call(self, unary, binary, index1, index2):
        """Join the unary and binary tensors.

        :param unary: the tensor with unary predicates pre-activations
        :param binary: the tensor with binary predicates pre-activations
        :param index1: a vector containing the indices of the first object
        of the pair referred by binary tensor
        :param index1: a vector containing the indices of the second object
        of the pair referred by binary tensor
        """

        index1 = tf.squeeze(index1)
        index2 = tf.squeeze(index2)

        # For the case where index1 and index2 were of length 1, tf.squeeze will make their rank = 0
        if index1.shape.rank == 0 and index2.shape.rank == 0:
            index1 = tf.reshape(index1, (1,))
            index2 = tf.reshape(index2, (1,))

        return tf.concat([tf.gather(unary, index1, axis=0), tf.gather(unary, index2, axis=0), binary], axis=1)



import tensorflow as tf


class GroupBy(tf.keras.layers.Layer):
    """GroupBy layer

    """

    def __init__(self, number_of_unary_predicates):
        """Initialize the GroupBy layer.

        """
        super(GroupBy, self).__init__()
        self.n_unary = number_of_unary_predicates

    def build(self, input_shape):
        """Build the layer

        :param input_shape: the input shape
        """
        super(GroupBy, self).build(input_shape)

    def call(self, unary, binary, deltas, index1, index2):
        """Split the deltas matrix in unary and binary deltas.

        :param unary: the tensor with unary predicates pre-activations
        :param binary: the tensor with binary predicates pre-activations
        :param deltas: the tensor containing the delta values
        :param index1: a vector containing the indices of the first object
        of the pair referred by binary and deltas tensors
        :param index2: a vector containing the indices of the second object
        of the pair referred by binary and deltas tensors
        """
        ux = deltas[:, :self.n_unary]
        uy = deltas[:, self.n_unary:2 * self.n_unary]
        b = deltas[:, 2 * self.n_unary:]
        shape = tf.shape(unary)

        return tf.scatter_nd(index1, ux, shape) + tf.scatter_nd(index2, uy, shape), b
    
